(** Camera of discardable fractions.

    This is a generalisation of the fractional camera where elements can
    represent both ownership of a fraction (as in the fractional camera) and the
    knowledge that a fraction has been discarded.

    Ownership of a fraction is denoted `DfracOwn q` and behave identically to
    `q` of the fractional camera.

    Knowledge that a fraction has been discarded is denoted `DfracDiscarded p`.
    Elements of this form are their own core, making ownership of them
    persistent. The operation combines knowledge that p and p' has been
    discarded into the knowledge that `p max p'` has been discarded.

    One can make a frame preserving update from _owning_ a fraction to _knowing_
    that the fraction has been discarded.

    Crucially, ownership over 1 is an exclusive element just as it is in the
    fractional camera. Hence owning 1 implies that no fraction has been
    discarded. Conversely, knowing that a fraction has been discarded implies
    that no one can own 1. And, since discarding is an irreversible operation it
    also implies that no one can own 1 in the future *)

From Coq.QArith Require Import Qcanon QArith_base Qminmax.
From iris.algebra Require Export cmra frac.
From iris.algebra Require Import proofmode_classes.
From iris.base_logic Require Import base_logic.
Set Default Proof Using "Type".

Open Scope Qc_scope.
Open Scope Qp_scope.

(* An element of dfrac denotes ownership of a fraction, or knowledge that a
   fraction has been discarded, or both. Note that all elements can be written
   in the form `DfracOwn q ⋅ DfracDiscarded p`, and that this should be
   preferred over using `DfracBoth`. *)
Inductive dfrac :=
  | DfracOwn : Qp → dfrac
  | DfracDiscarded : Qp → dfrac
  | DfracBoth : Qp → Qp → dfrac.

(* Projects the ownership part out of a dfrac. *)
Definition dfrac_own : dfrac -> option Qp := λ x,
  match x with
  | DfracOwn q => Some q
  | DfracDiscarded _ => None
  | DfracBoth q _ => Some q
  end.

(* Projects the discarded part out of a dfrac. *)
Definition dfrac_discarded : dfrac -> option Qp := λ x,
  match x with
  | DfracOwn _ => None
  | DfracDiscarded p => Some p
  | DfracBoth _ p => Some p
  end.

Section dfrac.
Canonical Structure dfracO := leibnizO dfrac.

(* An element is valid as long as the sum of its content is less than one. *)
Instance dfrac_valid : Valid dfrac := λ x,
  match x with
  | DfracOwn q => q ≤ 1%Qp
  | DfracDiscarded p => p ≤ 1%Qp
  | DfracBoth q p => (q + p)%Qp ≤ 1%Qp
  end.

(* As in the fractional camera the core is undefined for elements denoting
   ownership of a fraction. For elements denoting the knowledge that a fraction has
   been discarded the core is the identity function. *)
Instance dfrac_pcore : PCore dfrac := λ x, DfracDiscarded <$> (dfrac_discarded x).

(* When elements are combined, ownership is added together and knowledge of
   discarded fractions is combined with the max operation. *)
Instance dfrac_op : Op dfrac := λ x y,
  match x, y with
  | DfracOwn q, DfracOwn q' => DfracOwn (q + q')
  | DfracOwn q, DfracDiscarded p' => DfracBoth q p'
  | DfracOwn q, DfracBoth q' p' => DfracBoth (q + q') p'
  | DfracDiscarded p, DfracOwn q' => DfracBoth q' p
  | DfracDiscarded p, DfracDiscarded p' => DfracDiscarded (p `max` p')
  | DfracDiscarded p, DfracBoth q' p' => DfracBoth q' (p `max` p')
  | DfracBoth q p, DfracOwn q' => DfracBoth (q + q') p
  | DfracBoth q p, DfracDiscarded p' => DfracBoth q (p `max` p')
  | DfracBoth q p, DfracBoth q' p' => DfracBoth (q + q') (p `max` p')
  end.

Lemma dfrac_op_own (q p : Qp) : DfracOwn p ⋅ DfracOwn q = DfracOwn (p + q).
Proof. done. Qed.

Lemma dfrac_op_discarded (q p : Qp) :
  DfracDiscarded p ⋅ DfracDiscarded q = DfracDiscarded (p `max` q).
Proof. done. Qed.

Lemma dfrac_included_own (q p : Qp) : q < p ↔ DfracOwn q ≼ DfracOwn p.
Proof.
  rewrite Qp_lt_sum. split.
  - intros [z Heq]. exists (DfracOwn z). by rewrite dfrac_op_own Heq.
  - rewrite /included /op /dfrac_op. intros [[o|?|?] [= ->]]. by exists o.
Qed.

Lemma dfrac_included_discarded (q p : Qp) :
  q ≤ p ↔ DfracDiscarded q ≼ DfracDiscarded p.
Proof. 
  split.
  - intros ?. exists (DfracDiscarded p). by rewrite dfrac_op_discarded /Qp_max decide_True.
  - rewrite /included /op /dfrac_op. intros [[?|?|?] [= ->]]. apply Qc_le_max_l.
Qed.

Definition dfrac_ra_mixin : RAMixin dfrac.
Proof.
  split; try apply _.
  - intros [?|?|??] y cx <-; intros [= <-]; eexists _; done.
  - intros [?|?|??] [?|?|??] [?|?|??];
    try (rewrite /op /dfrac_op 1?Qp_plus_assoc 1?assoc; done).
  - intros [?|?|??] [?|?|??];
    rewrite /op /dfrac_op 1?Qp_plus_comm 1?(comm Qp_max); try done.
  - intros [?|?|??] cx; rewrite /pcore /dfrac_pcore; intros [= <-];
    rewrite /op /dfrac_op Qp_max_id; done.
  - intros [?|?|??] ? [= <-]; done.
  - intros [?|?|??] [?|?|??] ? [[?|?|??] [=]] [= <-]; eexists _; split; try done;
    apply dfrac_included_discarded; subst; auto; apply Qc_le_max_l.
  - intros [?|?|??] [?|?|??]; rewrite /op /dfrac_op /valid /dfrac_valid.
    * apply (Qp_plus_weak_r _ _ 1).
    * apply (Qp_plus_weak_r _ _ 1).
    * apply Qcle_trans. eapply Qcle_trans; last apply Qp_le_plus_l. apply Qp_le_plus_l.
    * apply (Qp_plus_weak_l _ _ 1).
    * apply (Qp_max_lub_l _ _ 1).
    * by intros ?%(Qp_plus_weak_l _ _ 1)%(Qp_max_lub_l _ _ 1).
    * rewrite (comm _ _ q1) -assoc. apply (Qp_plus_weak_l _ _ 1).
    * intros H. eapply Qcle_trans; last apply H.
      apply Qcplus_le_mono_l. apply Qc_le_max_l.
    * intros H. eapply Qcle_trans; last apply H.
      rewrite -assoc. apply Qcplus_le_mono_l, Qp_plus_weak_2_r, Qc_le_max_l.
Qed.
Canonical Structure dfracR := discreteR dfrac dfrac_ra_mixin.

Global Instance dfrac_cmra_discrete : CmraDiscrete dfracR.
Proof. apply discrete_cmra_discrete. Qed.
End dfrac.

Global Instance dfrac_full_exclusive : Exclusive (DfracOwn 1).
Proof.
  intros [?|?|??]; rewrite /op /cmra_op -cmra_discrete_valid_iff /valid /cmra_valid /=.
  - apply Qp_not_plus_q_ge_1.
  - apply Qp_not_plus_q_ge_1.
  - rewrite -Qcplus_assoc. apply (Qp_not_plus_q_ge_1 (q + q0)).
Qed.

Global Instance frac_cancelable (q : frac) : Cancelable q.
Proof. intros ?????. by apply Qp_eq, (inj (Qcplus q)), (Qp_eq (q+y) (q+z))%Qp. Qed.

Global Instance dfrac_discarded_core_id p : CoreId (DfracDiscarded p).
Proof. by constructor. Qed.

Lemma dfrac_valid_own (p : Qp) : ✓ DfracOwn p ↔ (p ≤ 1%Qp)%Qc.
Proof. done. Qed.

Lemma dfrac_valid_discarded (p : Qp) : ✓ DfracDiscarded p ↔ (p ≤ 1%Qp)%Qc.
Proof. done. Qed.

Lemma dfrac_valid_own_discarded (q p : Qp) :
  ✓ (DfracOwn q ⋅ DfracDiscarded p) ↔ (q + p ≤ 1%Qp)%Qc.
Proof. done. Qed.

Global Instance is_op_frac q : IsOp' (DfracOwn q) (DfracOwn (q/2)) (DfracOwn (q/2)).
Proof. by rewrite /IsOp' /IsOp dfrac_op_own Qp_div_2. Qed.

(* Discarding a fraction is a frame preserving update. *)
Lemma dfrac_discard_update q : DfracOwn q ~~> DfracDiscarded q.
  intros n [[?|?|??]|]; rewrite /op /cmra_op -!cmra_discrete_valid_iff /valid /cmra_valid /=.
  - by rewrite Qcplus_comm.
  - intro. eapply Qcle_trans. apply Qp_max_plus. done.
  - intro. eapply Qcle_trans; last done.
    rewrite -Qcplus_assoc. rewrite (Qcplus_comm q _). rewrite -Qcplus_assoc.
    apply Qcplus_le_mono_l. rewrite Qcplus_comm. apply Qp_max_plus.
  - done.
Qed.
