From iris.proofmode Require Import tactics.
From iris.algebra Require Export frac agree local_updates.
From iris.algebra Require Import proofmode_classes.
From iris.base_logic Require Import base_logic.
Set Default Proof Using "Type*".

(** PAuth is a generalization of the authorative RA where the type of the RA in
    the authorative and the fragmental view may differ. A projection function
    [π] converts from the type in the fragment to the type authorative. *)

Record pauth (A B : Type) (p : B → A) :=
  PAuth { pauth_auth_proj : option (frac * agree A) ; pauth_frag_proj : B }.
Add Printing Constructor pauth.
Arguments PAuth {_ _ _} _ _.
Arguments pauth_auth_proj {_ _ _} _.
Arguments pauth_frag_proj {_ _ _} _.
Instance: Params (@PAuth) 1 := {}.
Instance: Params (@pauth_auth_proj) 1 := {}.
Instance: Params (@pauth_frag_proj) 1 := {}.

Definition pauth_frag {A B : ucmraT} {π} (a: B) : pauth A B π := PAuth None a.
Definition pauth_auth {A B : ucmraT} {π} (q: Qp) (a: A) : pauth A B π :=
  PAuth (Some (q, to_agree a)) ε.

Typeclasses Opaque pauth_auth pauth_frag.

Instance: Params (@pauth_frag) 1 := {}.
Instance: Params (@pauth_auth) 1 := {}.

Notation "◯π a" := (pauth_frag a) (at level 20).
Notation "●π{ q } a" := (pauth_auth q a) (at level 20, format "●π{ q }  a").
Notation "●π a" := (pauth_auth 1 a) (at level 20).

(* Ofe *)
Section ofe.
Context {A B : ofeT}.
Context {π : B → A}.
Implicit Types a : option (frac * agree A).
Implicit Types b : B.
Implicit Types x y : pauth A B π.

Instance pauth_equiv : Equiv (pauth A B π) := λ x y,
  pauth_auth_proj x ≡ pauth_auth_proj y ∧ pauth_frag_proj x ≡ pauth_frag_proj y.
Instance pauth_dist : Dist (pauth A B π) := λ n x y,
  pauth_auth_proj x ≡{n}≡ pauth_auth_proj y ∧
  pauth_frag_proj x ≡{n}≡ pauth_frag_proj y.

Global Instance PAuth_ne : NonExpansive2 (@PAuth A B π).
Proof. by split. Qed.
Global Instance PAuth_proper : Proper ((≡) ==> (≡) ==> (≡)) (@PAuth A B π).
Proof. by split. Qed.
Global Instance pauth_auth_proj_ne: NonExpansive (@pauth_auth_proj A B π).
Proof. by destruct 1. Qed.
Global Instance pauth_auth_proj_proper : Proper ((≡) ==> (≡)) (@pauth_auth_proj A B π).
Proof. by destruct 1. Qed.
Global Instance pauth_frag_proj_ne : NonExpansive (@pauth_frag_proj A B π).
Proof. by destruct 1. Qed.
Global Instance pauth_frag_proj_proper : Proper ((≡) ==> (≡)) (@pauth_frag_proj A B π).
Proof. by destruct 1. Qed.

Definition pauth_ofe_mixin : OfeMixin (pauth A B π).
Proof. by apply (iso_ofe_mixin (λ x, (pauth_auth_proj x, pauth_frag_proj x))). Qed.
Canonical Structure pauthO := OfeT (pauth A B π) pauth_ofe_mixin.

Global Instance PAuth_discrete a b :
  Discrete a → Discrete b → Discrete (PAuth a b).
Proof. by intros ?? [??] [??]; split; apply: discrete. Qed.
Global Instance pauth_ofe_discrete : OfeDiscrete A → OfeDiscrete B → OfeDiscrete pauthO.
Proof. intros ?? [??]. apply _. Qed.
End ofe.

Arguments pauthO : clear implicits.

(* Camera *)
Section cmra.
Context {A B : ucmraT}.
Context (π : B → A).
Context `{!MonoidHomomorphism op op (≡) π}.

Instance pi_ne : NonExpansive π.
Proof. apply monoid_homomorphism_ne. Qed.
Instance: (Proper ((≡) ==> (≡)) π).
Proof. apply weak_monoid_homomorphism_proper. Qed.

Implicit Types a : A.
Implicit Types b : B.
Implicit Types x y : pauth A B π.

Global Instance pauth_frag_ne: NonExpansive (@pauth_frag A B π).
Proof. done. Qed.
Global Instance pauth_frag_proper : Proper ((≡) ==> (≡)) (@pauth_frag A B π).
Proof. done. Qed.
Global Instance pauth_auth_ne q : NonExpansive (@pauth_auth A B π q).
Proof. solve_proper. Qed.
Global Instance pauth_auth_proper : Proper ((≡) ==> (≡) ==> (≡)) (@pauth_auth A B π).
Proof. solve_proper. Qed.
Global Instance pauth_auth_discrete q a :
  Discrete a → Discrete (ε : B) → Discrete (●π{q} a : pauth A B π).
Proof. intros. apply PAuth_discrete; apply _. Qed.
Global Instance pauth_frag_discrete b : Discrete b → Discrete (◯π b : pauth A B π).
Proof. intros. apply PAuth_discrete; apply _. Qed.

Instance pauth_valid : Valid (pauth A B π) := λ x,
  match pauth_auth_proj x with
  | Some (q, ag) =>
      ✓ q ∧ (∀ n, ∃ a, ag ≡{n}≡ to_agree a ∧ π (pauth_frag_proj x) ≼{n} a ∧ ✓{n} a) ∧ ✓ pauth_frag_proj x
  | None => ✓ pauth_frag_proj x
  end.
Global Arguments pauth_valid !_ /.
Instance pauth_validN : ValidN (pauth A B π) := λ n x,
  match pauth_auth_proj x with
  | Some (q, ag) =>
      ✓{n} q ∧ ∃ a, ag ≡{n}≡ to_agree a ∧ π (pauth_frag_proj x) ≼{n} a ∧ ✓{n} a ∧ ✓{n} pauth_frag_proj x
  | None => ✓{n} pauth_frag_proj x
  end.
Global Arguments pauth_validN _ !_ /.
Instance pauth_pcore : PCore (pauth A B π) := λ x,
  Some (PAuth (core (pauth_auth_proj x)) (core (pauth_frag_proj x))).
Instance pauth_op : Op (pauth A B π) := λ x y,
  PAuth (pauth_auth_proj x ⋅ pauth_auth_proj y) (pauth_frag_proj x ⋅ pauth_frag_proj y).

Definition pauth_valid_eq :
  valid = λ x, match pauth_auth_proj x with
               | Some (q, ag) => ✓ q ∧ (∀ n, ∃ a, ag ≡{n}≡ to_agree a ∧ π (pauth_frag_proj x) ≼{n} a ∧ ✓{n} a) ∧ ✓ pauth_frag_proj x
               | None => ✓ pauth_frag_proj x
               end := eq_refl _.
Definition pauth_validN_eq :
  validN = λ n x, match pauth_auth_proj x with
                  | Some (q, ag) => ✓{n} q ∧ ∃ a, ag ≡{n}≡ to_agree a ∧ π (pauth_frag_proj x) ≼{n} a ∧ ✓{n} a ∧ ✓{n} pauth_frag_proj x
                  | None => ✓{n} pauth_frag_proj x
                  end := eq_refl _.

Lemma pauth_included (x y : pauth A B π) :
  x ≼ y ↔
  pauth_auth_proj x ≼ pauth_auth_proj y ∧ pauth_frag_proj x ≼ pauth_frag_proj y.
Proof.
  split; [intros [[z1 z2] Hz]; split; [exists z1|exists z2]; apply Hz|].
  intros [[z1 Hz1] [z2 Hz2]]; exists (PAuth z1 z2); split; auto.
Qed.

Lemma pauth_auth_proj_validN n x : ✓{n} x → ✓{n} pauth_auth_proj x.
Proof.
  destruct x as [[[]|]]; last done. intros (? & ? & Eq & ?). by rewrite Eq.
Qed.
Lemma pauth_frag_proj_validN n x : ✓{n} x → ✓{n} pauth_frag_proj x.
Proof.
  rewrite pauth_validN_eq.
  destruct x as [[[]|]]; [|done]. naive_solver eauto using cmra_validN_includedN.
Qed.
(* Lemma pauth_auth_proj_valid x : ✓ x → ✓ pauth_auth_proj x.
Proof.
  destruct x as [[[]|]]; [intros [? H]|done]. split; [done|].
  intros n. by destruct (H n) as [? [-> [? ?]]].
Qed. *)
Lemma pauth_frag_proj_valid x : ✓ x → ✓ pauth_frag_proj x.
Proof.
  destruct x as [[[]|]]; [by intros [_ [_ ?]]|done].
Qed.

Lemma pauth_frag_validN n b : ✓{n} (◯π b) ↔ ✓{n} b.
Proof. done. Qed.
(* Lemma pauth_auth_frac_validN n q a :
  ✓{n} (●{q} a) ↔ ✓{n} q ∧ ✓{n} a.
Proof.
  rewrite pauth_validN_eq /=. apply and_iff_compat_l. split.
  - by intros [?[->%(inj to_agree) []]].
  - naive_solver eauto using ucmra_unit_leastN.
Qed. *)
(* Lemma pauth_auth_validN n a : ✓{n} a ↔ ✓{n} (● a).
Proof.
  rewrite pauth_auth_frac_validN -cmra_discrete_valid_iff frac_valid'. naive_solver.
Qed. *)
Lemma pauth_both_frac_validN n q a b :
  ✓{n} (●π{q} a ⋅ ◯π b) ↔ ✓{n} q ∧ (π b) ≼{n} a ∧ ✓{n} a ∧ ✓{n} b.
Proof.
  rewrite pauth_validN_eq /=. apply and_iff_compat_l.
  setoid_rewrite (left_id _ _ b). split.
  - by intros (? & ->%(inj to_agree) & ?).
  - naive_solver.
Qed.
Lemma pauth_both_validN n a b : ✓{n} (●π a ⋅ ◯π b) ↔ (π b) ≼{n} a ∧ ✓{n} a ∧ ✓{n} b.
Proof.
  rewrite pauth_both_frac_validN -cmra_discrete_valid_iff frac_valid'. naive_solver.
Qed.

Lemma pauth_frag_valid b : ✓ (◯π b) ↔ ✓ b.
Proof. done. Qed.
Lemma pauth_auth_frac_valid q a : ✓ (●π{q} a) ↔ ✓ q ∧ ✓ a.
Proof.
  rewrite pauth_valid_eq /=. apply and_iff_compat_l. split.
  - intros [H' _]. apply cmra_valid_validN. intros n.
    by destruct (H' n) as (? & ->%(inj to_agree) & _ & h).
  - intros. split; last apply ucmra_unit_valid.
    exists a. split; first done.
    split; last by apply cmra_valid_validN.
    exists a. replace (ε : B) with monoid_unit by done.
    by rewrite monoid_homomorphism_unit left_id.
Qed.
Lemma pauth_auth_valid a : ✓ (●π a) ↔ ✓ a.
Proof. rewrite pauth_auth_frac_valid frac_valid'. naive_solver. Qed.

(* The reverse direction of the two lemmas below only holds if the camera is
discrete. *)
Lemma pauth_both_frac_valid_2 q a b : ✓ q → ✓ a → ✓ b → (π b) ≼ a → ✓ (●π{q} a ⋅ ◯π b).
Proof.
  assert (Proper ((≡) ==> (≡)) π). { apply weak_monoid_homomorphism_proper. }
  intros Val1 Val2 Val3 Incl. rewrite pauth_valid_eq /=.
  split; [done|]. split; [|by rewrite left_id].
  intros n. exists a. split; [done|]. 
  rewrite left_id.
  split; by [apply cmra_included_includedN|apply cmra_valid_validN].
Qed.
(* Lemma pauth_both_valid_2 a b : ✓ a → b ≼ a → ✓ (● a ⋅ ◯ b).
Proof. intros ??. by apply pauth_both_frac_valid_2. Qed. *)

Lemma pauth_valid_discrete `{!CmraDiscrete A, !CmraDiscrete B} x :
  ✓ x ↔ match pauth_auth_proj x with
        | Some (q, ag) => ✓ q ∧ ∃ a, ag ≡ to_agree a ∧ π (pauth_frag_proj x) ≼ a ∧ ✓ a ∧ ✓ (pauth_frag_proj x)
        | None => ✓ pauth_frag_proj x
        end.
Proof.
  rewrite pauth_valid_eq. destruct x as [[[??]|] ?]; simpl; [|done].
  setoid_rewrite <-cmra_discrete_included_iff.
  setoid_rewrite <-(discrete_iff _ a).
  setoid_rewrite <-cmra_discrete_valid_iff. naive_solver eauto using O.
Qed.
Lemma pauth_both_frac_valid `{!CmraDiscrete A, !CmraDiscrete B} q a b :
  ✓ (●π{q} a ⋅ ◯π b) ↔ ✓ q ∧ (π b) ≼ a ∧ ✓ a ∧ ✓ b.
Proof.
  rewrite pauth_valid_discrete /=. apply and_iff_compat_l.
  setoid_rewrite (left_id _ _ b). split.
  - by intros [?[->%(inj to_agree)]].
  - naive_solver.
Qed.
Lemma pauth_both_valid `{!CmraDiscrete A, !CmraDiscrete B} a b :
  ✓ (●π a ⋅ ◯π b) ↔ (π b) ≼ a ∧ ✓ a ∧ ✓ b.
Proof. rewrite pauth_both_frac_valid frac_valid'. naive_solver. Qed.

Lemma pauth_cmra_mixin : CmraMixin (pauth A B π).
Proof.
  apply cmra_total_mixin.
  - eauto.
  - by intros n x y1 y2 [Hy Hy']; split; simpl; rewrite ?Hy ?Hy'.
  - by intros n y1 y2 [Hy Hy']; split; simpl; rewrite ?Hy ?Hy'.
  - intros n [x a] [y b] [Hx Ha]; simpl in *. rewrite !pauth_validN_eq.
    destruct Hx as [x y Hx|]; first destruct Hx as [? Hx];
      [destruct x,y|]; intros VI; ofe_subst; auto.
  - intros [[[]|] ]; rewrite /= ?pauth_valid_eq ?pauth_validN_eq /=
      ?cmra_valid_validN; naive_solver.
  - intros n [[[]|] ]; rewrite !pauth_validN_eq /=;
      naive_solver eauto using dist_S, cmra_includedN_S, cmra_validN_S.
  - by split; simpl; rewrite assoc.
  - by split; simpl; rewrite comm.
  - by split; simpl; rewrite ?cmra_core_l.
  - by split; simpl; rewrite ?cmra_core_idemp.
  - intros ??; rewrite! pauth_included; intros [??].
    by split; simpl; apply: cmra_core_mono. (* FIXME: FIXME(Coq #6294): needs new unification *)
  - intros n [[[q1 a1]|] b1] [[[q2 a2]|] b2]; rewrite pauth_validN_eq /=;
      try naive_solver eauto using cmra_validN_op_l, cmra_validN_includedN.
    intros [? [a [Eq [Incl Valid]]]].
    assert (Eq': a1 ≡{n}≡ a2). { by apply agree_op_invN; rewrite Eq. }
    split; [naive_solver eauto using cmra_validN_op_l|].
    exists a. rewrite -Eq -Eq' agree_idemp. split; first done.
    pose proof Valid as [??]. split. move: Incl. rewrite monoid_homomorphism. apply cmra_includedN_weak_l.
     (* naive_solver. *)
    split; first done. eapply cmra_validN_op_l. done.
    intros [? [a [Eq [Incl Valid]]]].
    split; [naive_solver eauto using cmra_validN_op_l|].
    exists a. split; first done.
    pose proof Valid as [??]. split. move: Incl. rewrite monoid_homomorphism. apply cmra_includedN_weak_l.
    split; first done. eapply cmra_validN_op_l. done.
  - intros n x y1 y2 ? [??]; simpl in *.
    destruct (cmra_extend n (pauth_auth_proj x) (pauth_auth_proj y1)
      (pauth_auth_proj y2)) as (ea1&ea2&?&?&?); auto using pauth_auth_proj_validN.
    destruct (cmra_extend n (pauth_frag_proj x) (pauth_frag_proj y1) (pauth_frag_proj y2))
      as (b1&b2&?&?&?); auto using pauth_frag_proj_validN.
    by exists (PAuth ea1 b1), (PAuth ea2 b2).
Qed.
Canonical Structure pauthR := CmraT (pauth A B π) pauth_cmra_mixin.

Global Instance pauth_cmra_discrete : CmraDiscrete A → CmraDiscrete B → CmraDiscrete pauthR.
Proof.
  split; first apply _.
  intros [[[]|] ?]; rewrite pauth_valid_eq pauth_validN_eq /=; auto.
  - setoid_rewrite <-cmra_discrete_included_iff.
    rewrite -cmra_discrete_valid_iff.
    setoid_rewrite <-cmra_discrete_valid_iff.
    setoid_rewrite <-(discrete_iff _ a).
    intros [? Hv]. split; first done. destruct Hv as (a0 & HI & ? & ? & ?).
    split; last done. intros. exists a0. done.
  - by rewrite -cmra_discrete_valid_iff.
Qed.

Instance pauth_empty : Unit (pauth A B π) := PAuth ε ε.
Lemma pauth_ucmra_mixin : UcmraMixin (pauth A B π).
Proof.
  split; simpl.
  - rewrite pauth_valid_eq /=. apply ucmra_unit_valid.
  - by intros x; constructor; rewrite /= left_id.
  - do 2 constructor; [done| apply (core_id_core _)].
Qed.
Canonical Structure pauthUR := UcmraT (pauth A B π) pauth_ucmra_mixin.

Global Instance pauth_frag_core_id b : CoreId b → CoreId (◯π b).
Proof. do 2 constructor; simpl; auto. by apply core_id_core. Qed.

Lemma pauth_frag_op b b' : ◯π (b ⋅ b') = ◯π b ⋅ ◯π b'.
Proof. done. Qed.
Lemma pauth_frag_op' b b' : ◯π (b ⋅ b') = ◯π b ⋅ PAuth None b'.
Proof. done. Qed.
Lemma pauth_frag_op'' z b b' : PAuth (Some z) (b ⋅ b') = ◯π b ⋅ PAuth (Some z) b'.
Proof. done. Qed.
Lemma pauth_frag_mono b b' : b ≼ b' → ◯π b ≼ ◯π b'.
Proof. intros [c ->]. rewrite pauth_frag_op. apply cmra_included_l. Qed.
Lemma pauth_frag_core b : core (◯π b) = ◯π (core b).
Proof. done. Qed.

Lemma pauth_frag_update b b' : π b' ≼ π b → b ~~> b' → ◯π b ~~> ◯π b'.
Proof.
  intros Hle Hu.
  intros n [[[[q ?a]|] b'']|]; simpl. 
  - rewrite -pauth_frag_op'' pauth_validN_eq /=.
   intros (? & ? & ? & ? & ? & ?).
    split; first done. exists x. split; first done. split_and!.
    * rewrite monoid_homomorphism. setoid_rewrite monoid_homomorphism in H1. etrans.
      by apply cmra_monoN_r, cmra_included_includedN. done.
    * done.
    * by apply (Hu n (Some b'')).
  - rewrite -?pauth_frag_op'.
    intros Hv. epose proof (Hu n (Some b'')) as HI. simpl in HI. apply HI, Hv.
  - intros Hv. pose proof (Hu n ε) as HI. simpl in HI. apply HI, Hv.
Qed.

(*
Global Instance pauth_frag_sep_homomorphism :
  MonoidHomomorphism op op (≡) (@pauth_frag A B p).
Proof. by split; [split; try apply _|]. Qed.

Lemma big_opL_auth_frag {B} (g : nat → B → A) (l : list B) :
  (◯ [^op list] k↦x ∈ l, g k x) ≡ [^op list] k↦x ∈ l, ◯ (g k x).
Proof. apply (big_opL_commute _). Qed.
Lemma big_opM_auth_frag `{Countable K} {B} (g : K → B → A) (m : gmap K B) :
  (◯ [^op map] k↦x ∈ m, g k x) ≡ [^op map] k↦x ∈ m, ◯ (g k x).
Proof. apply (big_opM_commute _). Qed.
Lemma big_opS_auth_frag `{Countable B} (g : B → A) (X : gset B) :
  (◯ [^op set] x ∈ X, g x) ≡ [^op set] x ∈ X, ◯ (g x).
Proof. apply (big_opS_commute _). Qed.
Lemma big_opMS_auth_frag `{Countable B} (g : B → A) (X : gmultiset B) :
  (◯ [^op mset] x ∈ X, g x) ≡ [^op mset] x ∈ X, ◯ (g x).
Proof. apply (big_opMS_commute _). Qed.

Lemma pauth_both_op q a b : PAuth (Some (q,to_agree a)) b ≡ ●{q} a ⋅ ◯ b.
Proof. by rewrite /op /pauth_op /= left_id. Qed.

Lemma pauth_auth_frac_op p q a : ●{p + q} a ≡ ●{p} a ⋅ ●{q} a.
Proof.
  intros; split; simpl; last by rewrite left_id.
  by rewrite -Some_op -pair_op agree_idemp.
Qed.
Lemma pauth_auth_frac_op_invN n p a q b : ✓{n} (●{p} a ⋅ ●{q} b) → a ≡{n}≡ b.
Proof.
  rewrite /op /pauth_op /= left_id -Some_op -pair_op pauth_validN_eq /=.
  intros (?&?& Eq &?&?). apply (inj to_agree), agree_op_invN. by rewrite Eq.
Qed.
Lemma pauth_auth_frac_op_inv p a q b : ✓ (●{p} a ⋅ ●{q} b) → a ≡ b.
Proof.
  intros ?. apply equiv_dist. intros n.
  by eapply pauth_auth_frac_op_invN, cmra_valid_validN.
Qed.
Lemma pauth_auth_frac_op_invL `{!LeibnizEquiv A} q a p b :
  ✓ (●{p} a ⋅ ●{q} b) → a = b.
Proof. by intros ?%pauth_auth_frac_op_inv%leibniz_equiv. Qed.

(** Internalized properties *)
Lemma pauth_equivI {M} x y :
  x ≡ y ⊣⊢@{uPredI M} pauth_auth_proj x ≡ pauth_auth_proj y ∧ pauth_frag_proj x ≡ pauth_frag_proj y.
Proof. by uPred.unseal. Qed.
Lemma pauth_validI {M} x :
  ✓ x ⊣⊢@{uPredI M} match pauth_auth_proj x with
                    | Some (q, ag) => ✓ q ∧
                      ∃ a, ag ≡ to_agree a ∧ (∃ b, a ≡ pauth_frag_proj x ⋅ b) ∧ ✓ a
                    | None => ✓ pauth_frag_proj x
                    end.
Proof. uPred.unseal. by destruct x as [[[]|]]. Qed.
Lemma pauth_frag_validI {M} (a : A):
  ✓ (◯ a) ⊣⊢@{uPredI M} ✓ a.
Proof. by rewrite pauth_validI. Qed.

Lemma pauth_both_validI {M} q (a b: A) :
  ✓ (●{q} a ⋅ ◯ b) ⊣⊢@{uPredI M} ✓ q ∧ (∃ c, a ≡ b ⋅ c) ∧ ✓ a.
Proof.
  rewrite -pauth_both_op pauth_validI /=. apply bi.and_proper; [done|].
  setoid_rewrite agree_equivI. iSplit.
  - iDestruct 1 as (a') "[#Eq [H V]]". iDestruct "H" as (b') "H".
    iRewrite -"Eq" in "H". iRewrite -"Eq" in "V". auto.
  - iDestruct 1 as "[H V]". iExists a. auto.
Qed.
Lemma pauth_auth_validI {M} q (a b: A) :
  ✓ (●{q} a) ⊣⊢@{uPredI M} ✓ q ∧ ✓ a.
Proof.
  rewrite pauth_validI /=. apply bi.and_proper; [done|].
  setoid_rewrite agree_equivI. iSplit.
  - iDestruct 1 as (a') "[Eq [_ V]]". by iRewrite -"Eq" in "V".
  - iIntros "V". iExists a. iSplit; [done|]. iSplit; [|done]. iExists a.
    by rewrite left_id.
Qed.
*)

(* Definition splits b b1 b2 := b = b1 ⋅ b2 ∧ π b1 = ε ∧ π b2 = π b. *)

Lemma pauth_proj_unit : π ε ≡ ε.
Proof.
  replace (ε : B) with monoid_unit by done. by rewrite monoid_homomorphism_unit.
Qed.

Lemma pauth_update a b a' b' : 
  (∀ c n, ✓{n} (b ⋅ c) → (π b ⋅ π c ≼{n} a) → ✓{n} (b' ⋅ c) ∧ π (b' ⋅ c) ≼{n} a' ∧ ✓{n} a') →
  ●π a ⋅ ◯π b ~~> ●π a' ⋅ ◯π b'.
Proof.
  intros Hip. apply cmra_total_update.
  move=> n [[qa|] bf1] [/= Hq [a0 [Haa0 [Ha [Ha0 Hv']]]]]; do 2 red; simpl in *.
  (* move=> n [[qa|] bf1] [/= Hq [a0 [Haa0 [[bf2 Ha] [Ha0 Hv']]]]]; do 2 red; simpl in *. *)
  { exfalso. eapply exclusiveN_l; last apply Hq. apply _. }
  apply (inj to_agree) in Haa0.
  split; first done. exists a'. split; first done.
  destruct (Hip bf1 n) as (? & ? & ?).
  - move: Hv'. rewrite left_id. done.
  - move: Ha. rewrite left_id. rewrite Haa0. rewrite monoid_homomorphism. done.
  - split_and!; rewrite ?left_id; done.
Qed.

Lemma pauth_update_local a b a' b' :
  (a,π b) ~l~> (a',π b') →
  (∀ c n, ✓{n} (b ⋅ c) → (π b ⋅ π c ≼{n} a) → ✓{n} (π b' ⋅ π c) → ✓{n} (b' ⋅ c)) →
  ●π a ⋅ ◯π b ~~> ●π a' ⋅ ◯π b'.
Proof.
  intros Hup Hv. apply cmra_total_update.
  move=> n [[qa|] bf1] [/= Hq [a0 [Haa0 [[bf2 Ha] [Ha0 Hv']]]]]; do 2 red; simpl in *.
  { by apply (exclusiveN_l _ _) in Hq. }
  split; [done|]. apply (inj to_agree) in Haa0.
  move: Ha. rewrite !left_id monoid_homomorphism -assoc=> Ha.
  destruct (Hup n (Some (π bf1 ⋅ bf2))); [by rewrite Haa0|by rewrite Haa0|]; simpl in *.
  exists a'. split_and!; [done| |done|].
  exists bf2. by rewrite left_id monoid_homomorphism -assoc.
  rewrite left_id.
  move: Hv'. rewrite left_id. intro Hv'.
  apply Hv. done.
  - move: Ha. rewrite -Haa0 assoc. by eexists _.
  - move: H. rewrite H0 assoc. intro H. apply cmra_validN_op_l in H. done.
Qed.

Lemma pauth_update_alloc a a' b' :
  (a,ε) ~l~> (a',π b') →
  (∀ c n, ✓{n} c → π c ≼{n} a → ✓{n} (π b' ⋅ π c) → ✓{n} (b' ⋅ c)) →
  ●π a ~~> ●π a' ⋅ ◯π b'.
Proof.
  intros ? H. rewrite -(right_id _ _ (●π a)). apply pauth_update_local;
  first by rewrite pauth_proj_unit.
  intros x n. rewrite pauth_proj_unit. rewrite !left_id. apply H.
Qed.

Lemma pauth_update_alloc_alt a a' b' pb' :
  (a,ε) ~l~> (a',pb') →
  π b' = pb' →
  (∀ c n, ✓{n} c → π c ≼{n} a → ✓{n} (pb' ⋅ π c) → ✓{n} (b' ⋅ c)) →
  ●π a ~~> ●π a' ⋅ ◯π b'.
Proof. intros H Eq. move: H. rewrite -Eq. apply pauth_update_alloc. Qed.

(* Lemma pauth_update_auth a a' c : (a,ε) ~l~> (a',c) → ●π a ~~> ●π a'.
Proof.
  intros. etrans; first exact: pauth_update_alloc.
  exact: cmra_update_op_l.
Qed.

Lemma pauth_update_core_id q a b `{!CoreId b} :
  b ≼ a → ●{q} a ~~> ●{q} a ⋅ ◯ b.
Proof.
  intros Ha%(core_id_extract _ _). apply cmra_total_update.
  move=> n [[[q' ag]|] bf1] [Hq [a0 [Haa0 [Hbf1 Ha0]]]]; do 2 red; simpl in *.
  + split; [done|]. exists a0. split_and!; [done| |done].
    assert (a ≡{n}≡ a0) as Haa0' by (apply to_agree_includedN; by exists ag).
    rewrite -Haa0' Ha Haa0' -assoc (comm _ b) assoc. by apply cmra_monoN_r.
  + split; [done|]. exists a0. split_and!; [done| |done].
    apply (inj to_agree) in Haa0.
    rewrite -Haa0 Ha Haa0 -assoc (comm _ b) assoc. by apply cmra_monoN_r.
Qed.

Lemma pauth_update_dealloc a b a' : (a,b) ~l~> (a',ε) → ● a ⋅ ◯ b ~~> ● a'.
Proof. intros. rewrite -(right_id _ _ (● a')). by apply pauth_update. Qed.

Lemma pauth_local_update (a b0 b1 a' b0' b1': A) :
  (b0, b1) ~l~> (b0', b1') → b0' ≼ a' → ✓ a' →
  (● a ⋅ ◯ b0, ● a ⋅ ◯ b1) ~l~> (● a' ⋅ ◯ b0', ● a' ⋅ ◯ b1').
Proof.
  rewrite !local_update_unital=> Hup ? ? n /=.
    move=> [[[qc ac]|] bc] /pauth_both_validN [Le Val] [] /=.
  - move => Ha. exfalso. move : Ha. rewrite right_id -Some_op -pair_op frac_op'.
    move => /Some_dist_inj [/= Eq _].
    apply (Qp_not_plus_q_ge_1 qc). by rewrite -Eq.
  - move => _. rewrite !left_id=> ?.
    destruct (Hup n bc) as [Hval' Heq]; eauto using cmra_validN_includedN.
    rewrite -!pauth_both_op pauth_validN_eq /=.
    split_and!; [done| |done]. exists a'.
    split_and!; [done|by apply cmra_included_includedN|by apply cmra_valid_validN].
Qed.
*)
End cmra.

(* Arguments pauthR : clear implicits. *)
Arguments pauthR _ _ _ {_}.
Arguments pauthUR _ _ _ {_}.

(*
(* Proof mode class instances *)
Instance is_op_auth_frag {A : ucmraT} (a b1 b2 : A) :
  IsOp a b1 b2 → IsOp' (◯ a) (◯ b1) (◯ b2).
Proof. done. Qed.
Instance is_op_auth_auth_frac {A : ucmraT} (q q1 q2 : frac) (a : A) :
  IsOp q q1 q2 → IsOp' (●{q} a) (●{q1} a) (●{q2} a).
Proof. rewrite /IsOp' /IsOp => ->. by rewrite -pauth_auth_frac_op. Qed.

(* Functor *)
Definition pauth_map {A B} (f : A → B) (x : pauth A) : pauth B :=
  PAuth (prod_map id (agree_map f) <$> pauth_auth_proj x) (f (pauth_frag_proj x)).
Lemma pauth_map_id {A} (x : pauth A) : pauth_map id x = x.
Proof. destruct x as [[[]|] ]; by rewrite // /pauth_map /= agree_map_id. Qed.
Lemma pauth_map_compose {A B C} (f : A → B) (g : B → C) (x : pauth A) :
  pauth_map (g ∘ f) x = pauth_map g (pauth_map f x).
Proof. destruct x as [[[]|] ];  by rewrite // /pauth_map /= agree_map_compose. Qed.
Lemma pauth_map_ext {A B : ofeT} (f g : A → B) `{!NonExpansive f} x :
  (∀ x, f x ≡ g x) → pauth_map f x ≡ pauth_map g x.
Proof.
  constructor; simpl; auto.
  apply option_fmap_equiv_ext=> a; by rewrite /prod_map /= agree_map_ext.
Qed.
Instance pauth_map_ne {A B : ofeT} (f : A -> B) `{Hf : !NonExpansive f} :
  NonExpansive (pauth_map f).
Proof.
  intros n [??] [??] [??]; split; simpl in *; [|by apply Hf].
  apply option_fmap_ne; [|done]=> x y ?. apply prod_map_ne; [done| |done].
  by apply agree_map_ne.
Qed.
Instance pauth_map_cmra_morphism {A B : ucmraT} (f : A → B) :
  CmraMorphism f → CmraMorphism (pauth_map f).
Proof.
  split; try apply _.
  - intros n [[[??]|] b]; rewrite !pauth_validN_eq;
      [|naive_solver eauto using cmra_morphism_monotoneN, cmra_morphism_validN].
    intros [? [a' [Eq [? ?]]]]. split; [done|]. exists (f a').
    rewrite -agree_map_to_agree -Eq.
    naive_solver eauto using cmra_morphism_monotoneN, cmra_morphism_validN.
  - intros [??]. apply Some_proper; rewrite /pauth_map /=.
    by f_equiv; rewrite /= cmra_morphism_core.
  - intros [[[??]|]?] [[[??]|]?]; try apply PAuth_proper=>//=; by rewrite cmra_morphism_op.
Qed.
Definition pauthO_map {A B} (f : A -n> B) : pauthO A -n> pauthO B :=
  OfeMor (pauth_map f).
Lemma pauthO_map_ne A B : NonExpansive (@pauthO_map A B).
Proof. intros n f f' Hf [[[]|] ]; repeat constructor; try naive_solver;
  apply agreeO_map_ne; auto. Qed.

Program Definition pauthRF (F : urFunctor) : rFunctor := {|
  rFunctor_car A _ B _ := pauthR (urFunctor_car F A B);
  rFunctor_map A1 _ A2 _ B1 _ B2 _ fg := pauthO_map (urFunctor_map F fg)
|}.
Next Obligation.
  by intros F A1 ? A2 ? B1 ? B2 ? n f g Hfg; apply pauthO_map_ne, urFunctor_map_ne.
Qed.
Next Obligation.
  intros F A ? B ? x. rewrite /= -{2}(pauth_map_id x).
  apply (pauth_map_ext _ _)=>y; apply urFunctor_map_id.
Qed.
Next Obligation.
  intros F A1 ? A2 ? A3 ? B1 ? B2 ? B3 ? f g f' g' x. rewrite /= -pauth_map_compose.
  apply (pauth_map_ext _ _)=>y; apply urFunctor_map_compose.
Qed.

Instance pauthRF_contractive F :
  urFunctorContractive F → rFunctorContractive (pauthRF F).
Proof.
  by intros ? A1 ? A2 ? B1 ? B2 ? n f g Hfg; apply pauthO_map_ne, urFunctor_map_contractive.
Qed.

Program Definition pauthURF (F : urFunctor) : urFunctor := {|
  urFunctor_car A _ B _ := pauthUR (urFunctor_car F A B);
  urFunctor_map A1 _ A2 _ B1 _ B2 _ fg := pauthO_map (urFunctor_map F fg)
|}.
Next Obligation.
  by intros F A1 ? A2 ? B1 ? B2 ? n f g Hfg; apply pauthO_map_ne, urFunctor_map_ne.
Qed.
Next Obligation.
  intros F A ? B ? x. rewrite /= -{2}(pauth_map_id x).
  apply (pauth_map_ext _ _)=>y; apply urFunctor_map_id.
Qed.
Next Obligation.
  intros F A1 ? A2 ? A3 ? B1 ? B2 ? B3 ? f g f' g' x. rewrite /= -pauth_map_compose.
  apply (pauth_map_ext _ _)=>y; apply urFunctor_map_compose.
Qed.

Instance pauthURF_contractive F :
  urFunctorContractive F → urFunctorContractive (pauthURF F).
Proof.
  by intros ? A1 ? A2 ? B1 ? B2 ? n f g Hfg; apply pauthO_map_ne, urFunctor_map_contractive.
Qed.
*)